/**
  * hgtfp version 1
  *
  * Implementation of the HGT/FP algorithm (Harmonic Greedy Triplets with the Four-Point Condition).
  *
  * Please send me a note about your experience with this program, especially with 
  * non-artificial data sets. 
  * 
  * Reference:
  * [Csu02] Miklos Csuros. Fast recovery of evolutionary trees 
  *         with thousands of nodes. Journal of Computational Biology,
  *         9(2):277-297. 2002.
  *
  * Related publications:
  * [Csu00a] Miklos Csuros. Fast recovery of evolutionary trees with 
  *         thousands on nodes. RECOMB 2001, 104113 [abstract for Csu02]. 
  * [CsKa02] Miklos Csuros and Ming-Yang Kao. Provably fast and accurate 
  *         recovery of evolutionary trees through Harmonic Greedy Triplets. 
  *         SIAM Journal on Computing, 31:306322, 2001.
  * [CsKa99] Miklos Csuros and Ming-Yang Kao. Recovering evolutionary trees 
  *         through Harmonic Greedy Triplets. SODA 1999, 261270 [abstract for CsKa02].240-453-3324
  * [Csu00b] Miklos Csuros. Reconstructing Phylogenies in Markov Models of Evolution. 
  *         Ph. D. dissertation, Department of Computer Science, Yale University, 2000.
  *         UMI Publication Number: AAT 9991139 [http://wwwlib.umi.com/dissertations/]. 
  *
  * (You can find preprints of these publications on my webpage.)
  *
  *
  * Miklos Csuros, Department of Computer Science and Operations Research, University of Montreal.
  * csuros@iro.umontreal.ca http://www.iro.umontreal.ca/~csuros/
  *
  *******************************************************************************************
  * I/O
  *
  * if input is not given, then standard input is read [Phylip's symmetric distance matrix format].
  * output is written onto stdout [Newick format w/ possible newlines]
  *
  ******************************************************************************************
  * 
  * Compiling & running
  *
  * $(CC) -o hgtfp hgtfp.c -lm
  *
  * where $(CC) is your C compiler: gcc is recommended; gcc version 2.95 is tested: no warnings with -Wall -ansi
  * 
  * Running: 
  * hgtfp <first taxon> [<Phylip format distance file>] 
  *
  *
  ******************************************************************************************
  * 
  * Distribution and copyright notice
  *
  * Copyright (C) 2002 Miklos Csuros
  *
  * This program is free software; you can redistribute it and/or
  * modify it under the terms of the GNU General Public License
  * version 2 as published by the Free Software Foundation.
  *
  * This program is distributed in the hope that it will be useful,
  * but WITHOUT ANY WARRANTY; without even the implied warranty of
  * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  * GNU General Public License for more details.
  *
  * You can find the GNU General Public License at http://www.gnu.org/licenses/gpl.html.
  * For a hardcopy, write to the Free Software
  * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
  *
  * In addition to the free distribution of the source file, you are granted permission to
  * run the compiled executable for evaluation, academic, and other non-commercial purposes.
  *    
  ******************************************************************************************
  *   
  * $Id: hgtfp.c,v 1.1 2002/05/08 17:36:04 csuros Exp csuros $
  *
  */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <limits.h>
#include <ctype.h>

#define Float double
#define Bool short
#define iNode int /* indexing method for nodes: 1...N: iNode==0 is used to signal special conditions [errors, defaults, etc.] */ 
#define Dist Float
#define True 1
#define False 0

/* BIG_DBL is unused now, it can be used for encoding "infinity" in the input matrix 
   by values larger than BIG_DBL, and then set infinite distance in ReadMatrix() to
   the largest finite value: Tandy Warnow's suggestion for dealing with infinite distances b/w taxa */
#define BIG_DBL 1e10

/*
  data structure for tree nodes: in addition to the usual binary tree data structure 
  fields, the defining triplet is stored with directional information
*/
typedef struct {
  char * name;
  iNode parent; /* parent node */
  iNode left; /* left child */
  iNode right;  /* right child */

  Dist length; /* edge length */
  
  int color; /* node coloring for temporary stuff */
  
  iNode def_up; /* defining triplet */
  iNode def_left;
  iNode def_right;

  Bool in_tree; /* if added to the tree yet */ 
  Bool isRoot; /* = (parent != 0) */
  Bool isLeaf; /* = (left != 0 || right != 0); */
} Node;

/* test: does the defining triplet for P include X? */
#define isDefBy(P, X) ((P).def_left == (X) || (P).def_right == (X) || (P).def_up == (X))

typedef struct {
  iNode N; /* triplet NXY */
  iNode X;
  iNode Y;

  Dist D_P1; /* new edge lengths */
  Dist D_P2;
  Dist D_N;

  Float score; /* average similarity */
  iNode P2; /* edge on which this triplet is relevant: edges are referred to by the child node */
} Split;

/* array of relevant triplets */
typedef Split* Splitarray;

/* in order to die "gracefully" when there isn't enough memory available */ 
#define safe_calloc(var,size,typ) {if ((var = (typ *) calloc(size,sizeof(typ))) == NULL) {fprintf(stderr, "Out of memory in line %d\n", __LINE__); exit(1);}}

#define _upper_half_of_matrix(m, i, j) ((i)>(j) ? (m[i][j]): (m[j][i]))
#define _lower_half_of_matrix(m, i, j) ((j)>(i) ? (m[i][j]): (m[j][i]))

/* use the same matrix for storing distances and similarities in upper and lower halves 
 * "similarity" is used by Csu02, Csu00a, and Csu00b, 
 * "closeness" is used by CsKa99 and CsKa02: they refer to the same thing: e^{-distance}
 *
 */ 
#define Distance(data, i, j)   _upper_half_of_matrix(data, i, j)
#define Closeness(data, i,j)   _lower_half_of_matrix(data, i, j)
#define CloseTriplet(data, i, j, k) (3.0/(1.0/Closeness(data, i, j)+1.0/Closeness(data, i, k)+1.0/Closeness(data, j, k)))

/* main subroutines */
int BuildTree(
   Node*,     /* array of tree nodes: 1-based array */
   Float **,  /* array of distances */
   int ,
   iNode );

int  SplitEdge();
/* return values from SplitEdge() */
#define SPLIT_YES      1
#define SPLIT_NO       0

/* auxiliary subroutines: for detailed discussion, see my dissertation [Csu00b]  */
Bool  StartingTriplet();
void UpdateRelevant();
void AddNewTuples();

void InitializeTree();
void AddLeaf();

void   TupleUpdate();
Split* TupleBest();
void   TupleDelEdge();

/* I/O */
Float ** ReadMatrix();
void PrintTree();

#ifndef PRINT_TREE_NODES
/* max. number of nodes per line: 
   dumb programs [and Newick standard] want the whole tree in one line, 
   I prefer listing only a handful nodes per line to avoid really long lines 
   with large trees  */
#define PRINT_TREE_NODES 1e10 
#endif

#ifndef TAXON_NAME_LEN
/* max. length of taxon names: could do a spiffier, length-independent name parsing 
   but this should be ok also - who wants more than 25 lines of taxon name? */
#define TAXON_NAME_LEN 2048 
#endif

/* define VERBOSE with -DVERBOSE at compile time for debug messages */
/*
#define VERBOSE
*/

int main(int argc, char **argv){
  iNode first_taxon;
  char *file_name;
  char *stdin_name = "-";
  Float **Data;
  char  **taxon_name;
  int num_leaves;
  int i;

  Node *Tree;

  if (argc<2 || argc>3){
    printf("Usage: %s <first taxon number> [<distance file in Phylip format>]\n",
	   argv[0]);
    exit(1);
  }
  first_taxon = atoi(argv[1]);
  file_name = (argc == 3 ? argv[2] : stdin_name);
  
  Data = ReadMatrix(file_name, &num_leaves, &taxon_name);

  safe_calloc(Tree, 2*num_leaves-1, Node);
  for (i=1; i<=num_leaves; i++) Tree[i].name=taxon_name[i];
  
  if (BuildTree(Tree, Data, num_leaves, first_taxon))
    PrintTree(Tree);
  
  for(i=1; i<=num_leaves; i++){
    free(Data[i]);
    if (Tree[i].name) free(Tree[i].name);
  }
  free(Tree);
  free(Data);

  return 0;
}


/* ------------------------ BuildTree -------------------- */

int BuildTree(Node T[], Float **Data, 
	      int num_leaves,
	      iNode first_taxon){
  /** 
   *
   *  T[1..num_leaves] has the terminal taxa, with the name field set
   *  T[num_leaves+1...2*num_leaves-2] are the internal nodes
   *
   * returns the number of taxa added in the tree
   */
  
  Splitarray split;
  int n;
  iNode X,Y,Z;
  Bool found;
  iNode N, P, P2;
  Split* best_split;
  int i;

#ifdef VERBOSE
  printf("#*BT num_leaves %d first %d\n", num_leaves, first_taxon);
#endif  
  
  found = StartingTriplet(&X, &Y, &Z, Data, num_leaves, first_taxon, first_taxon);
  
  if (!found){
    fputs("No starting triplet found.", stderr);
    return 0;
  }
#ifdef VERBOSE
  printf("#*BT first triplet X %d Y %d Z %d\n", X, Y, Z);
#endif
  
  /* initialize the tree as the triplet formed by XYZ: X is the root */
  InitializeTree(T, num_leaves, X, Y, Z, 
		 Distance(Data, X, Y),
		 Distance(Data, X, Z),
		 Distance(Data, Y, Z));
  n = 4; /* returns n-1 */
  /* allocate space for the splits */

  safe_calloc(split, num_leaves+1, Split);
  
  /* initialize the splits */
  for (i=1; i<=num_leaves; i++) split[i].P2 = 0;

  /* check all the relevant triplets */
  P = num_leaves+1; /* index of the internal node */
  for(N=1; N<=num_leaves; N++) 
    if (N != X && N != Y && N != Z){
      TupleUpdate(T, Data, split, P, N, X, Y);
      TupleUpdate(T, Data, split, Y, N, X, Y);
      TupleUpdate(T, Data, split, P, N, X, Z);
      TupleUpdate(T, Data, split, Z, N, X, Z);
      TupleUpdate(T, Data, split, Y, N, Y, Z);
      TupleUpdate(T, Data, split, Z, N, Y, Z);      
    }

  /* add the remaining taxa */
  for(n=4; n<=num_leaves; n++){ 
    best_split = TupleBest(split, num_leaves);
    if (best_split == NULL){
      fprintf(stderr, "Cannot find suitable triplet for the %d-th taxon\n", n);
      break;
    }
    P2 = best_split->P2; 
    P++;
    AddLeaf(T, P,
	    best_split->N, best_split->P2,
	    best_split->D_P1, best_split->D_P2, best_split->D_N,
	    best_split->X, best_split->Y);
    if ( n < num_leaves){ /* update relevant triplets */
      TupleDelEdge(split, P2, num_leaves);
      UpdateRelevant(T, Data, split, P2, num_leaves);
    }  	 
  } /* for n */
  
  if (split) free(split);
  return(n-1); 
}

/* ------------------------ SplitEdge -------------------- */

int SplitEdge(Dist *D_1, Dist *D_2, Dist *D_N,
	      Node T[], Float **Data, 
	      iNode P2, iNode N, iNode X, iNode Y){
  /** sets D_1, D_2, D_N if returns SPLIT_YES
    *  tree T[], distance/closeness Data[],
    *  edge P2, triplet NXY
    */
  Dist D_XN = Distance(Data, X, N);
  Dist D_YN = Distance(Data, Y, N);
  Dist D_XY = Distance(Data, X, Y);
  Dist D_XP = (D_XN+D_XY-D_YN)/2.0;
  Dist D_YP = (D_YN+D_XY-D_XN)/2.0;
  Dist D_NP = (D_XN+D_YN-D_XY)/2.0;
  Dist D1, D2;

  iNode P1=T[P2].parent;
  
  Dist d_xy, d_xz, d_yz, d;
  iNode U;

  int quart_ok_up, quart_total_up, quart_ok_dn, quart_total_dn;
  Dist D_long1, D_long2, D_long3, D_long4, D_short1, D_short2;
  quart_total_up=quart_total_dn=0;
  
  /* 
   * -------------------------------------- 
   * calculate edge lengths to use when NXY is added
   *
   */
  if (T[P1].isRoot){
    D1 = (X == P1 ? D_XP : D_YP);

    quart_total_up=0;
    quart_ok_up=0;
  }
  else { /* P1 is an internal node */
    d_xy = Distance(Data, T[P1].def_up, T[P1].def_left);
    d_xz = Distance(Data, T[P1].def_up, T[P1].def_right);
    d_yz = Distance(Data, T[P1].def_left, T[P1].def_right);
    if (     T[P1].def_up    == X){ U=X; d = (d_xy+d_xz-d_yz)/2.0; }
    else if (T[P1].def_up    == Y){ U=Y; d = (d_xy+d_xz-d_yz)/2.0; }
    else if (T[P1].def_left  == X){ U=X; d = (d_xy+d_yz-d_xz)/2.0; }
    else if (T[P1].def_left  == Y){ U=Y; d = (d_xy+d_yz-d_xz)/2.0; }
    else if (T[P1].def_right == X){ U=X; d = (d_xz+d_yz-d_xy)/2.0; }
    else if (T[P1].def_right == Y){ U=Y; d = (d_xz+d_yz-d_xy)/2.0; }
    else { /* bad call: shouldn't ever happen */
      fprintf(stderr, "SplitEdge: P1 does not share a leaf with XYN in line %d\n", __LINE__);
      exit(9);
    }
    D1 = (U == X ? D_XP : D_YP)-d;
    if ((P2 == T[P1].left && U == T[P1].def_left)
	|| (P2 == T[P1].right && U == T[P1].def_right)) D1 = -D1;

#ifdef VERBOSE
    printf("#*SE '%d,%d', P1 %d X %d Y %d N %d D_XY %g D_XN %g D_YN %g D_XP %g D_YP %g dxy %g dxz %g dyz %g U %d d %g D1 %g\n",
	   P2, N, P1, X, Y, N, D_XY, D_XN, D_YN, D_XP, D_YP, d_xy, d_xz, d_yz, U, d, D1);
#endif
    
  /* 
   * -------------------------------------- 
   * got edge lengths in D1, D2
   *
   * now check if NXY can be added 
   *
   */
    
    /* find out how many quartets support the relation to P1 */
    quart_total_up=2;
    quart_ok_up=0;
    if (P2 == T[P1].left){
      if (X==T[P1].def_up || X==T[P1].def_right) {
	D_short1 = d_xz; /* up <-> right */
	D_short2 = Distance(Data, T[P1].def_left, N);
	D_long1  = d_xy; /* up <-> left */
	D_long2  = d_yz; /* left <-> right */
	D_long3  = Distance(Data, T[P1].def_up, N);
	D_long4  = Distance(Data, T[P1].def_right, N);
	
	if (D_long1+D_long4 > D_short1+D_short2) quart_ok_up ++ ;
	if (D_long2+D_long3 > D_short1+D_short2) quart_ok_up ++ ;
      } else if (Y==T[P1].def_up || Y==T[P1].def_right) {
	D_short1 = d_xz; /* up <-> right */
	D_short2 = Distance(Data, T[P1].def_left, N);
	D_long1  = d_xy; /* up <-> left */
	D_long2  = d_yz; /* left <-> right */
	D_long3  = Distance(Data, T[P1].def_up, N);
	D_long4  = Distance(Data, T[P1].def_right, N);
	
	if (D_long1+D_long4 > D_short1+D_short2) quart_ok_up ++ ;
	if (D_long2+D_long3 > D_short1+D_short2) quart_ok_up ++ ;
      } else {
	D_short1 = d_xz;
	D_short2 = Distance(Data, T[P1].def_left, N);
	D_long1  = d_xy; /* up <-> left */
	D_long2  = d_yz; /* left <-> right */
	D_long3  = Distance(Data, T[P1].def_up, N);
	D_long4  = Distance(Data, T[P1].def_right, N);
	
	if (D_long1+D_long4 > D_short1+D_short2) quart_ok_up ++ ;
	if (D_long2+D_long3 > D_short1+D_short2) quart_ok_up ++ ;      
	
	/*printf("#*SE left %d\n", quart_ok_up);*/
      }
    } else { /* P2 is on the right */
      if (X==T[P1].def_up || X==T[P1].def_left) {
	D_short1 = d_xy; /* up <-> left */
	D_short2 = Distance(Data, T[P1].def_right, N);
	D_long1  = d_xz; /* up <-> right */
	D_long2  = d_yz; /* left <-> right */
	D_long3  = Distance(Data, T[P1].def_up, N);
	D_long4  = Distance(Data, T[P1].def_left, N);
	
	if (D_long1+D_long4 > D_short1+D_short2) quart_ok_up ++ ;
	if (D_long2+D_long3 > D_short1+D_short2) quart_ok_up ++ ;
      } else if (Y==T[P1].def_up || Y==T[P1].def_left) {
	D_short1 = d_xy; /* up <-> left */
	D_short2 = Distance(Data, T[P1].def_right, N);
	D_long1  = d_xz; /* up <-> right */
	D_long2  = d_yz; /* left <-> right */
	D_long3  = Distance(Data, T[P1].def_up, N);
	D_long4  = Distance(Data, T[P1].def_left, N);
	
	if (D_long1+D_long4 > D_short1+D_short2) quart_ok_up ++ ;
	if (D_long2+D_long3 > D_short1+D_short2) quart_ok_up ++ ;
      } else {
	D_short1 = d_xz;
	D_short2 = Distance(Data, T[P1].def_left, N);
	D_long1  = d_xy; /* up <-> left */
	D_long2  = d_yz; /* left <-> right */
	D_long3  = Distance(Data, T[P1].def_up, N);
	D_long4  = Distance(Data, T[P1].def_right, N);
	
	if (D_long1+D_long4 > D_short1+D_short2) quart_ok_up ++ ;
	if (D_long2+D_long3 > D_short1+D_short2) quart_ok_up ++ ;      
	
	/*printf("#*SE right %d\n", quart_ok_up);*/
      }
    } /* if P2's direction */
  } /* if P1 is internal */
  
  if (T[P2].isLeaf){
    D2 = (X == P2 ? D_XP : D_YP);
    quart_ok_dn=0;
    quart_total_dn=0;
  } else { /* P2 is an internal node */
    d_xy = Distance(Data, T[P2].def_up, T[P2].def_left);
    d_xz = Distance(Data, T[P2].def_up, T[P2].def_right);
    d_yz = Distance(Data, T[P2].def_left, T[P2].def_right);
    if (     T[P2].def_up    == X){ U=X; d = (d_xy+d_xz-d_yz)/2.0; }
    else if (T[P2].def_up    == Y){ U=Y; d = (d_xy+d_xz-d_yz)/2.0; }
    else if (T[P2].def_left  == X){ U=X; d = (d_xy+d_yz-d_xz)/2.0; }
    else if (T[P2].def_left  == Y){ U=Y; d = (d_xy+d_yz-d_xz)/2.0; }
    else if (T[P2].def_right == X){ U=X; d = (d_xz+d_yz-d_xy)/2.0; }
    else if (T[P2].def_right == Y){ U=Y; d = (d_xz+d_yz-d_xy)/2.0; }
    else {
      fprintf(stderr, "SplitEdge: P2 does not share a leaf with XYN in line %d\n", __LINE__);
      exit(9);
    }
    D2 = (U == X ? D_XP : D_YP)-d;
    if (U == T[P2].def_up) D2 = -D2;
#ifdef VERBOSE
    printf("#*SE '%d,%d', P2 %d X %d Y %d N %d D_XY %g D_XN %g D_YN %g D_XP %g D_YP %g dxy %g dxz %g dyz %g U %d d %g D2 %g\n",
	   P2, N, P2, X, Y, N, D_XY, D_XN, D_YN, D_XP, D_YP, d_xy, d_xz, d_yz, U, d, D2);
#endif
    
    quart_total_dn=2;
    quart_ok_dn=0;
    if (X==T[P2].def_left || X==T[P2].def_right){
      D_short1 = d_yz; /* left <-> right */
      D_short2 = Distance(Data, T[P2].def_up, N);
      D_long1  = d_xy; /* up <-> left */
      D_long2  = d_xz; /* up <-> right */
      D_long3  = Distance(Data, N, T[P2].def_left);
      D_long4  = Distance(Data, N, T[P2].def_right);
      
      if (D_long1+D_long4 > D_short1+D_short2) quart_ok_dn ++ ;
      if (D_long2+D_long3 > D_short1+D_short2) quart_ok_dn ++ ;
    } else if (Y==T[P2].def_left || Y==T[P2].def_right) {
      D_short1 = d_yz; /* left <-> right */
      D_short2 = Distance(Data, T[P2].def_up, N);
      D_long1  = d_xy; /* up <-> left */
      D_long2  = d_xz; /* up <-> right */
      D_long3  = Distance(Data, N, T[P2].def_left);
      D_long4  = Distance(Data, N, T[P2].def_right);
      
      if (D_long1+D_long4 > D_short1+D_short2) quart_ok_dn ++ ;
      if (D_long2+D_long3 > D_short1+D_short2) quart_ok_dn ++ ;
    } else {
      D_short1 = d_yz; /* left <-> right */
      D_short2 = Distance(Data, T[P2].def_up, N);
      D_long1  = d_xy; /* up <-> left */
      D_long2  = d_xz; /* up <-> right */
      D_long3  = Distance(Data, N, T[P2].def_left);
      D_long4  = Distance(Data, N, T[P2].def_right);
      if (D_long1+D_long4 > D_short1+D_short2) quart_ok_dn ++ ;
      if (D_long2+D_long3 > D_short1+D_short2) quart_ok_dn ++ ;
      /*printf("#*SE up %d\n", quart_ok_dn);*/
    }
  } /* if P2 is internal */
#ifdef VERBOSE
  printf("#*SE X %d Y %d N %d edge %d:%d len %g D1 %g D2 %g\n",
	 X, Y, N, P2, P1, T[P2].length, D1, D2);
#endif

#ifdef VERBOSE
  printf("##*SE X %d Y %d N %d edge %d:%d q_up %d/%d q_dn:%d/%d %s %s \n",
	 X, Y, N, P2, P1, quart_ok_up, quart_total_up, quart_ok_dn, quart_total_dn,
	 (quart_total_up>0 ? (quart_ok_up<quart_total_up ? "---" : "+++") : "   "),
	 (quart_total_dn>0 ? (quart_ok_dn<quart_total_dn ? "---" : "+++") : "   "));
#endif
  if (quart_ok_dn < quart_total_dn || quart_ok_up<quart_total_up) return SPLIT_NO;
  
  *D_1 = (D1+T[P2].length-D2)/2.0;
  *D_2 = (D2+T[P2].length-D1)/2.0;
  *D_N = D_NP;
  
  return SPLIT_YES;
}

/* ------------------------ StartingTriplet -------------------- */

Bool StartingTriplet(iNode *X, iNode *Y, iNode *Z,
		     Float **Data,
		     int num_leaves,
		     iNode start_taxon, iNode end_taxon){
  /** returns 0 if no starting triplet found
   *  sets starting triplet in XYZ
   *  distance/closeness Data[],
   *  X taken from the range start_taxon...end_taxon
   *
   */
  
  iNode i, j, k;
  Float c_max, c_ij, c_ik, c_jk, c_ijk;
  Bool success = False;
  
  c_max = 0.0; 
  for(i=start_taxon; i<=end_taxon; i++)
    for (j = (start_taxon == 1? i+1: 1); j<num_leaves; j++){
      c_ij = (i!=j ? Closeness(Data, i, j) : 0.0);
      if (c_ij > c_max/3.0) /* b/c to beat c_max, all of them must be > */
	for (k = j+1; k<=num_leaves; k++){
	  c_ik = (i!=k ? Closeness(Data, i, k): 0.0);
	  c_jk = Closeness(Data, j, k);
	  if (c_ik > c_max/3.0 && c_jk > c_max/3.0){
	    c_ijk = 3.0/(1.0/c_ij+1.0/c_ik+1.0/c_jk);
	    if (c_ijk > c_max){
	      c_max = c_ijk;
	      success = True;
	      *X = i; *Y = j; *Z = k;
	    }
	  }
	} /* if, for k */
    } /* for i, j */
  return success;
}

/* ------------------------ UpdateRelevant -------------------- */

void UpdateRelevant(Node T[], Float **Data, Splitarray S,
		    iNode P2, int num_leaves){
  /** Called when a new node [parent of P2] is added
   *  to check all the new relevant triplets
   *
   */
#define __UR_AUX(p,x,y) AddNewTuples(T, Data, S, num_leaves, p, x, y)
  iNode P1, P, N;
  Bool P_on_left, P_def_dn, 
    P_shares_up_1, P_shares_dn_1, P_shares_up_2, P_shares_dn_2;

  /* the new internal node P and the new leaf N
   *   were just added on the edge def'd by P2 
   */
  P = T[P2].parent;
  P1 = T[P].parent; /* P1->P->P2 are all on the left or all on the right*/
  P_on_left = (T[P].left == P2);
  if (P_on_left){
    N = T[P].right;
    P_def_dn = T[P].def_left;
  }
  else {
    N = T[P].left;
    P_def_dn = T[P].def_right; 	/* P is def'd by P_def_up, P_def_dn, and N */
  }
#ifdef VERBOSE
  printf("#*UR P2 %d P1 %d P %d N %d\n", P2, P1, P, N);
#endif
  
  P_shares_up_1 = isDefBy(T[P1], T[P].def_up);
  P_shares_dn_1 = isDefBy(T[P1], P_def_dn);
  
  P_shares_up_2 = isDefBy(T[P2], T[P].def_up);
  P_shares_dn_2 = isDefBy(T[P2], P_def_dn);
  
  /* relevant triples share different definining leaves on endpoints */
  
  /* tuples on edge N:P */
  __UR_AUX(N, N, T[P].def_up);
  __UR_AUX(N, N, P_def_dn);
  
  /* on edge P:P1 */
  if (T[P1].isRoot){
    __UR_AUX(P, P_def_dn, P1);
    __UR_AUX(P, N, P1);
  }
  else {
    __UR_AUX(P, P_def_dn, T[P1].def_up);
    __UR_AUX(P, N, T[P1].def_up);
    if (P_on_left){
      __UR_AUX(P, P_def_dn, T[P1].def_right);
      __UR_AUX(P, N, T[P1].def_right);
      if (!P_shares_up_1 || !P_shares_dn_1) 
	__UR_AUX(P, T[P].def_up, T[P1].def_left);
    }
    else { /* P on right */
      __UR_AUX(P, P_def_dn, T[P1].def_left);
      __UR_AUX(P, N, T[P1].def_left);
      if (!P_shares_up_1 || !P_shares_dn_1) 
	__UR_AUX(P, T[P].def_up, T[P1].def_right);
    }
  }
  
  /* on edge P2:P */
  if (T[P2].isLeaf){
    __UR_AUX(P2, N, P2);
    __UR_AUX(P2, T[P].def_up, P2);
  }
  else {
    __UR_AUX(P2, N, T[P2].def_left);
    __UR_AUX(P2, N, T[P2].def_right);
    __UR_AUX(P2, T[P].def_up, T[P2].def_right);
    __UR_AUX(P2, T[P].def_up, T[P2].def_left);
    if (!P_shares_up_2 || !P_shares_dn_2)
      __UR_AUX(P2, P_def_dn, T[P2].def_up);
  }
  /* max. 12 calls to AddNewTuples */
  
}

/* ------------------------ AddNewTuples -------------------- */


void AddNewTuples(Node T[], Float **Data, Splitarray S,
		  int num_leaves, 
		  iNode P, iNode X, iNode Y){
  /** Checks all relevant tuples of the form NXY wrt the
   *  edge b/w P and its parent 
   *
   * [calls TupleUpdate with N set to all leaves not in tree yet]
   * 
   */
  iNode N;

#ifdef VERBOSE
  printf("#*AN P %d X %d Y %d\n", P, X, Y);
#endif
  for(N=1; N<=num_leaves; N++)
    if (T[N].in_tree == 0) /* N is not in tree yet */
      TupleUpdate(T, Data, S, P, N, X, Y);
}

/* ------------------------ InitializeTree -------------------- */


void InitializeTree(Node T[], int num_leaves,
		    iNode X, iNode Y, iNode Z,
		    Dist D_XY, Dist D_XZ, Dist D_YZ){
  /** tree T[],
   *  number of taxa num_leaves
   *  starting triplet XYZ
   *
   */
   
  iNode i;
#ifdef VERBOSE
  printf("#*IT numl %d X %d Y %d Z %d D_XY %g D_XZ %g D_YZ %g\n",
	 num_leaves, X, Y, Z, D_XY, D_XZ, D_YZ);
#endif

  for(i=1; i<=2*num_leaves-2; i++) {
    T[i].parent = T[i].left = T[i].right 
      = T[i].def_up = T[i].def_left = T[i].def_right = 0;
    T[i].in_tree = T[i].isLeaf = T[i].isRoot = False;
  }
  
  T[X].in_tree = True;
  T[X].left = Y;
  T[X].def_up = X;
  T[X].isRoot = True;
  T[X].isLeaf = False;
  
  T[Y].in_tree = True;
  T[Y].parent = X;
  T[Y].def_left = Y;
  T[Y].isLeaf = True;
  T[Y].isRoot = False;
  
  AddLeaf(T, num_leaves+1, Z, Y, 
  				(D_XY+D_XZ-D_YZ)/2.0, /* D_XP */
  				(D_XY+D_YZ-D_XZ)/2.0, /* D_YP */
  				(D_XZ+D_YZ-D_XY)/2.0, /* D_ZP */
  				X, Y); 
}

/* ------------------------ AddLeaf -------------------- */

void AddLeaf(Node T[], iNode P,
	     iNode N, iNode P2, 
	     Dist D_P1, Dist D_P2, Dist D_N,
	     iNode X, iNode Y){
  /** tree T[],
    *  new internal node index P,
    *  P added as parent of P2
    *  triplet NXY
    *
    * Figures out the right topology to set def_up, def_left and def_right
    *
    * Keeps the original left-right directions of the edge
    * (ie if P2 is on left edge, then N is a right child,
    *  otherwise N is a left child)
    */
#define __ADDLEAF_AUX(u, l, r) {T[P].def_up = u; T[P].def_left = l; T[P].def_right = r; }
  iNode P1 = T[P2].parent;

#ifdef VERBOSE
    printf("#*AL P %d N %d X %d Y %d edge %d:%d\n", P, N, X, Y, P2, P1);
#endif

  T[P].in_tree = True;
  T[P].parent = P1;
  T[P].length = D_P1;
  T[P].isRoot = T[P].isLeaf = False;
  
  T[N].in_tree = True;
  T[N].parent = P;
  T[N].length = D_N;
  T[N].def_left = N;
  T[N].isLeaf = True;
  T[N].isRoot = False;
  
  T[P2].parent = P;
  T[P2].length = D_P2;
  
  if (T[P1].left == P2){ /* insert on left edge */
    T[P].left = P2;
    T[P].right = N;
    T[P1].left = P;
    if (T[P2].isLeaf){
      if (X == P2){
	__ADDLEAF_AUX( Y, X, N);
      }
      else if (Y == P2){
	__ADDLEAF_AUX( X, Y, N);
      }
      else { /* shouldn't happen */
	fprintf(stderr, "AddLeaf: P2 is a leaf but != X, Y in line %d\n", __LINE__);
	exit(21);
      }
    }
    else { /* P2 is an internal node */
      if (X == T[P2].def_up){
	__ADDLEAF_AUX( X, Y, N);
      }
      else if (Y == T[P2].def_up){
	__ADDLEAF_AUX( Y, X, N);
      }
      else if (X == T[P2].def_left || X == T[P2].def_right){
	__ADDLEAF_AUX( Y, X, N);
      }
      else if (Y == T[P2].def_left || Y == T[P2].def_right){
	__ADDLEAF_AUX( X, Y, N);
      }
      else { /* shouldn't happen */
	fprintf(stderr, "AddLeaf: P2 shares no leaf w/ X, Y in line %d\n", __LINE__);
	exit(21);
      }
    }
  }
  else { /* insert on right edge */
    T[P].right = P2;
    T[P].left = N;
    T[P1].right = P;
    if (T[P2].isLeaf){
      if (X == P2){
	__ADDLEAF_AUX( Y, N, X);
      }
      else if (Y == P2){
	__ADDLEAF_AUX( X, N, Y);
      }
      else { /* shouldn't happen */
	fprintf(stderr, "AddLeaf: P2 is a leaf but != X, Y in line %d\n", __LINE__);
	exit(21);
      }
    }
    else { /* P2 is an internal node */
      if (X == T[P2].def_up){
	__ADDLEAF_AUX( X, N, Y);
      }
      else if (Y == T[P2].def_up){
	__ADDLEAF_AUX( Y, N, X);
      }
      else if (X == T[P2].def_left || X == T[P2].def_right){
	__ADDLEAF_AUX( Y, N, X);
      }
      else if (Y == T[P2].def_left || Y == T[P2].def_right){
	__ADDLEAF_AUX( X, N, Y);
      }
      else { /* shouldn't happen */
	fprintf(stderr, "AddLeaf: P2 shares no leaf w/ X, Y in line %d\n", __LINE__);
	exit(21);
      }
    }
  }
}

/* ------------------------ TupleUpdate -------------------- */

void TupleUpdate(Node T[], Float **Data, Splitarray S,
		 iNode P2, iNode N, iNode X, iNode Y){
  /** tree T[], distance/closeness Data[]
   *  splits S[], triplet NXY for edge P2P1
   *
   * Updates the split set if SplitEdge okays it *and* the
   * score of the triplet is larger than what we had before [for the greedy selection] 
   *
   */
  Float c_XY = Closeness(Data, X, Y);
  Float c_XN = Closeness(Data, X, N);
  Float c_YN = Closeness(Data, Y, N);
  Float avg_close;
  Split *entry;
  int split_edge;
  Dist D_P1, D_P2, D_N;
  
  if (c_XY<=0.0 || c_XN<=0.0 || c_YN<=0.0) return; /* not a positive triplet */
  avg_close = 3.0/(1.0/c_XY+1.0/c_XN+1.0/c_YN);

  entry = S+N;

  if (entry->P2 == 0 || entry->score<avg_close){ /* may be a better splitting triplet */
    if ((split_edge
	 =SplitEdge(&D_P1, &D_P2, &D_N, T, Data, 
		    P2, N, X, Y)) == SPLIT_YES){
      entry->N = N;
      entry->X = X;
      entry->Y = Y;
      entry->D_P1 = D_P1;
      entry->D_P2 = D_P2;
      entry->D_N = D_N;
      entry->score = avg_close;
      entry->P2 = P2;
#ifdef VERBOSE
      printf("#*TU new  X %d Y %d N %d e %d:%d\n", X, Y, N, P2, T[P2].parent);
#endif 
    }
    else {
#ifdef VERBOSE
      printf("#*TU bad  X %d Y %d N %d e %d:%d\n", X, Y, N, P2, T[P2].parent);
#endif
    }
  }
  else {
#ifdef VERBOSE
    printf("#*TU stay X %d Y %d N %d e %d:%d avg %g max %g X %d Y %d N %d e %d:%d]\n",
	   X, Y, N, P2, T[P2].parent, avg_close, entry->score,
	   entry->X, entry->Y, entry->N, entry->P2, T[entry->P2].parent);
#endif
  }
}
 
/* ------------------------ TupleBest -------------------- */

Split* TupleBest(Splitarray S, int num_leaves){
	Split *best = NULL;
	Float best_score = 0.0;
	iNode i;
	Split *this;

	for (i=1; i<=num_leaves; i++){
 		this = S+i;
		if (this->P2 &&  this->score > best_score){
			best = this;
			best_score = this->score;
		}
	}
	return best;
}

/* ------------------------ TupleDelEdge -------------------- */

void TupleDelEdge(Splitarray S, int P2, int num_leaves){
  iNode i;
  for (i=1; i<=num_leaves; i++)
    if (S[i].P2 == P2) S[i].P2 = 0;
}

/* ------------------------ ReadMatrix -------------------- 
  reads Phylip-format input distance matix
*/
Float** ReadMatrix(const char *filename, int *num_leaves, char*** taxon_names){
  FILE *input;
  Float **data;
  char **taxon;
  int i, j;
  char c;
  
  if (strcmp(filename, "-")==0){ /* input from stdin */
    input = stdin;
  }
  else {
    input = fopen(filename, "r");
    if (input == NULL){
      fprintf(stderr, "File %s not found.\n", filename);
      exit(2);
    }
  };
  
  *num_leaves = 0;
  fscanf(input, "%d", num_leaves);
  if (*num_leaves < 3){
    fprintf(stderr, "Phylip distance file must start with the number of taxa [>=3]\n");
    exit(3);
  }
  safe_calloc(taxon, *num_leaves+1, char*);
  for (i=1; i<=*num_leaves; i++) safe_calloc(taxon[i], TAXON_NAME_LEN+1, char);
  *taxon_names = taxon;
  safe_calloc(data, *num_leaves+1, Float*);
  for (i=1; i<=*num_leaves; i++) safe_calloc(data[i], *num_leaves+1, Float);
  
  for (i=1; i<=*num_leaves; i++){
#ifdef VERBOSE
    printf("#*RM taxon %d/%d\n", i, *num_leaves);
#endif
   
    do { 
      c = getc(input);
    } while (isspace(c));
    ungetc(c, input);
    if (c == '\'' || c == '\"'){ /* some taxon name */
      for(j=0; j<TAXON_NAME_LEN; j++){
	taxon[i][j] = getc(input);
	if (taxon[i][j] == c && j>0) break;
      }
      if (taxon[i][j] != c){
	fprintf(stderr, "Data file format error: cannot find ending quote for %d-th taxon name\n", i);
	exit(4);
      }
      taxon[i][j+1] = '\0';
    }
    else 
      if (fscanf(input,"%s", taxon[i])!= 1){
      fprintf(stderr, "Data file format error reading %d-th taxon name\n.", i);
      exit(4);
      };
    for(j=1; j<=*num_leaves; j++){
      if (fscanf(input, "%lg", data[i]+j) != 1){
	fprintf(stderr, "Data file format error reading distance[%d][%d]\n.", 
		i, j);
	exit(5);
      }
      /* printf("#*RM read d[%d][%d] = %g\n", i, j, data[i][j]);*/
    }
  }
  
  if (strcmp(filename, "-")!=0) fclose(input);

  /* get rid of BIG values: if uncommented, then selects bigdist, the maximum distance in the matrix
     that's below "infinity" (BIG_DBL) and sets every "infinity" to that value */
  /*
  {
    Float bigdist=0.;
  for (i=1; i<=*num_leaves; i++)
    for (j=1; j<=*num_leaves; j++)
     if (data[i][j] < BIG_DBL && data[i][j] > bigdist) bigdist = data[i][j];
  for (i=1; i<=*num_leaves; i++)
    for (j=1; j<=*num_leaves; j++)
      if (data[i][j] >= BIG_DBL) data[i][j] = bigdist;
  }
  */
  
  /* symmetrize the matrix and compute closeness */
  for (i=1; i<*num_leaves; i++){
    for (j=i+1; j<=*num_leaves; j++){
      /* printf("#*RM ij %g ji %g\n", data[i][j], data[j][i]);*/
      /* gcc permits conditional expression as lvalue, but ANSI C doesn't: 
         in the following two lines, array indexes are given directly on the lhs */ 
      /*Distance(data, i, j)*/ data[j][i] = (data[i][j]+data[j][i])/2.0;
      /*Closeness(data, i, j)*/ data[i][j] = exp(-Distance(data, i, j));
      /* printf("#*RM i %d j %d D %g c %g\n", i, j, Distance(data, i, j), Closeness(data, i, j));*/
    };
    data[i][i] = 0.0;
  };
  return data;
}
/* ------------------------ PrintTree -------------------- */
void _PrintSubtree(Node T[], iNode P){
  static int in_this_line; 
  if (T==NULL){ in_this_line = 0; return; };
  if (T[P].isLeaf){
    printf("%s:%g ", T[P].name, T[P].length);
    in_this_line++;
  }
  else {
    printf("(");
    if (T[P].left) _PrintSubtree(T, T[P].left);
    if (T[P].left && T[P].right){
      printf(", ");
      if (in_this_line >= PRINT_TREE_NODES){
	printf("\n ");
	in_this_line = 0;
      };
    };
    if (T[P].right) _PrintSubtree(T, T[P].right);
    printf("):%g", T[P].length); 
    if (in_this_line >= PRINT_TREE_NODES){
      printf("\n ");
      in_this_line = 0;
    };
  }
}

void PrintTree(Node T[]){
  iNode P;
  /* find root */
  for (P=1; !T[P].isRoot; P=T[P].parent);
  _PrintSubtree(NULL, 0); /* initialize */
  printf("(%s:%g, ", T[P].name, T[T[P].left].length);
  P = T[P].left;
  _PrintSubtree(T, T[P].left);
  printf(", ");
  _PrintSubtree(T, T[P].right);
  printf("):0;\n");
}
 
/*

  Revision 1.2  1999/09/05 19:14:28  csm
  working version

  Revision 1.1  1999/07/19 15:24:07  csm
  Initial revision

  Revision 1.6  1999/04/16 04:07:31  csm
  incorporated min.evol heuristic & fixed taxon name parsing

*/
